unit KpiSong;
{
 * NokomoDecoder
 * Copyright (c) 2005 Autch.net. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 *     Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 *     Neither the name of the Autch.net nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}

interface

uses Windows, SysUtils, Classes, KPI, kpiWrapper, CriticalSection;

type
	// ȃt@CƁCɊ֘AtꂽvOC̃bp[
	// TAuKPIDriver ɍ点̂]܂B
	TAuSong = class(TObject)
		private
			function GetSeekable: boolean;
			function GetLoopFlag: boolean;
			function GetMultiSongFlag: boolean;
			procedure SetPosition(const dwPos: DWord);
			function GetPosition: DWord;

			function SeekTo(const dwPos: DWord): DWord;
			procedure OpenFromFile(const filename: TFileName);
			procedure OpenFromBuffer(const Buffer: PByte; const dwSize: DWord);
			function BytesToMSecs(const dwBytes: DWord): DWord;
			function MSecsToBytes(const dwMSecs: DWord): DWord;
			function BytesPerSecond: DWord;
			procedure SetEndFlag;
			procedure ResetEndFlag;
			function GetPluginFileName: TFileName;
			procedure SetSongNum(const NewSongNum: DWord);
			function GetSoundInfo: PSOUNDINFO;
		protected
			m_Plugin: TAuKpi;
			m_Buffer: PByte;
			m_dwBufSize: DWord;
			m_FileName: TFileName;
			m_SoundInfo: SOUNDINFO;
			m_dwPosition: DWord;			// in bytes
			m_hKMP: HKMP;
			m_CriticalSection: TAuCriticalSection;
			m_EndFlag: Boolean;
			m_dwSongNum: DWord;

			FOnSongEnd: TNotifyEvent;
		public
			constructor Create(const filename: TFileName; const plugin: TAuKpi); overload;
			constructor Create(const filename: TFileName; const plugin: TAuKpi; const pInfoDesired: PSOUNDINFO); overload;
			constructor Create(const Buffer: PByte; const dwSize: DWord; const plugin: TAuKpi); overload;
			constructor Create(const Buffer: PByte; const dwSize: DWord; const plugin: TAuKpi; const pInfoDesired: PSOUNDINFO); overload;
			destructor Destroy; override;
			procedure Close;
			function Render(const Buffer: PByte; const dwSize: DWord): DWord; overload;

			property SamplesPerSec: DWord read m_SoundInfo.dwSamplesPerSec;
			property Channels: DWord 			read m_SoundInfo.dwChannels;
			property BitsPerSample: DWord read m_SoundInfo.dwBitsPerSample;
			property Length: DWord 				read m_SoundInfo.dwLength;
			property Seekable: boolean 		read GetSeekable;
			property UnitRender: DWord 		read m_SoundInfo.dwUnitRender;
			property Loop: Boolean				read GetLoopFlag;
			property MultiSong: Boolean		read GetMultiSongFlag;
			property Position: DWord			read GetPosition write SetPosition;		// in milliseconds
			property EndFlag: Boolean			read m_EndFlag;
			property PluginFileName: TFileName read GetPluginFileName;
			property SongNumber: DWord		read m_dwSongNum write SetSongNum;
			property SoundInfo: PSOUNDINFO read GetSoundInfo;

			property OnSongEnd: TNotifyEvent read FOnSongEnd write FOnSongEnd;
	end;

implementation

uses Math;

// w肳ꂽvOCgăt@CJ
constructor TAuSong.Create(const filename: TFileName; const plugin: TAuKpi);
begin
	inherited Create;
  m_CriticalSection := TAuCriticalSection.Create;
	m_Plugin := plugin;
  OpenFromFile(filename);
end;

// w肳ꂽvOCgăt@CJFPCM `w肷
constructor TAuSong.Create(const filename: TFileName; const plugin: TAuKpi; const pInfoDesired: PSOUNDINFO);
begin
	inherited Create;
  m_CriticalSection := TAuCriticalSection.Create;
	m_Plugin := plugin;
	m_SoundInfo := pInfoDesired^;
	OpenFromFile(filename);
end;

// w肳ꂽvOCgăobt@JFgp
constructor TAuSong.Create(const Buffer: PByte; const dwSize: DWord; const plugin: TAuKpi);
begin
	inherited Create;
	m_CriticalSection := TAuCriticalSection.Create;
	m_Plugin := plugin;
	OpenFromBuffer(Buffer, dwSize);
end;

// w肳ꂽvOC PCM `găobt@JFgp
constructor TAuSong.Create(const Buffer: PByte; const dwSize: DWord; const plugin: TAuKpi; const pInfoDesired: PSOUNDINFO);
begin
	inherited Create;
	m_CriticalSection := TAuCriticalSection.Create;
	m_Plugin := plugin;
	m_SoundInfo := pInfoDesired^;
  OpenFromBuffer(Buffer, dwSize);
end;

destructor TAuSong.Destroy;
begin
	if(m_hKMP <> nil) then Close;
  if(m_Buffer <> nil) then FreeMem(m_Buffer);
  m_CriticalSection.Free;
	// m_Plugin ͂ł Free ȂiTAuKpi ̏L҂ TAuKpiDriverj
	inherited;
end;

procedure TAuSong.OpenFromFile(const filename: TFileName);
begin
	m_FileName := filename;
	m_hKMP := m_Plugin.Open(m_FileName, @m_SoundInfo);
	ResetEndFlag;
end;

procedure TAuSong.OpenFromBuffer(const Buffer: PByte; const dwSize: DWord);
begin
	GetMem(m_Buffer, dwSize);
	MoveMemory(m_Buffer, Buffer, dwSize);
	m_dwBufSize := dwSize;
	m_hKMP := m_Plugin.OpenFromBuffer(m_Buffer, m_dwBufSize, @m_SoundInfo);
	ResetEndFlag;
end;

procedure TAuSong.Close;
begin
	m_Plugin.Close(m_HKMP);
	m_HKMP := nil;
end;

function TAuSong.GetSeekable: boolean;
begin
	Result := m_SoundInfo.dwSeekable;
end;

// bN͕KvȂ
// bN悤ȃvOC͎łĂ͂
function TAuSong.Render(const Buffer: PByte; const dwSize: DWord): DWord;
begin
//	m_CriticalSection.Lock;
	try
		Result := m_Plugin.Render(m_hKMP, Buffer, dwSize);
		Inc(m_dwPosition, Result);
		if (Result < dwSize) then SetEndFlag;
	finally
	//	m_CriticalSection.UnLock;
	end;
end;

function TAuSong.GetLoopFlag: boolean;
begin
	Result := m_SoundInfo.dwLoopFlag;
end;

function TAuSong.GetMultiSongFlag: boolean;
begin
	Result := m_SoundInfo.dwMultiSongFlag;
end;

procedure TAuSong.SetPosition(const dwPos: DWord);
begin
	m_dwPosition := SeekTo(dwPos);
end;

function TAuSong.GetPosition: DWord;
begin
	Result := BytesToMSecs(m_dwPosition);
end;

// vOCV[NT|[gĂȂĂV[N
// dwPos  milliseconds
function TAuSong.SeekTo(const dwPos: DWord): DWord;
var
	dwRet: DWord;
	bufDummy: PByte;
	dwRenderedSize: DWord;
begin
	//m_CriticalSection.Lock;
	ResetEndFlag;
	try
		Result := 0;
		if (m_SoundInfo.dwSeekable = true) then
		begin
			// V[NłȂb͊ȒP
			dwRet := m_plugin.SetPosition(m_hKMP, dwPos);
			Result := MSecsToBytes(dwRet);
		end
		else
		begin
			// V[NłȂƂ́CȂ̓_~[ Render() JԂē
			GetMem(bufDummy, m_SoundInfo.dwUnitRender);
			try
				m_Plugin.SetPosition(m_hKMP, 0);
				dwRet := 0;
				while (Abs(dwRet - MSecsToBytes(dwPos)) < Integer(m_SoundInfo.dwUnitRender)) do
				begin
					dwRenderedSize := m_Plugin.Render(m_hKMP, bufDummy, m_SoundInfo.dwUnitRender);
					Inc(dwRet, dwRenderedSize);
					if (dwRenderedSize < m_SoundInfo.dwUnitRender) then
					begin
						SetEndFlag;
						break;
					end;
				end;
				Result := dwRet;
			finally
				FreeMem(bufDummy);
			end;
		end;
	finally
		//m_CriticalSection.UnLock;
	end;
end;

// PʊZFb̃oCg
function TAuSong.BytesPerSecond: DWord;
begin
	with m_SoundInfo do
		Result := dwSamplesPerSec * (dwBitsPerSample div 8) * dwChannels;
end;

// PʊZFoCg~b
function TAuSong.BytesToMSecs(const dwBytes: DWord): DWord;
begin
	Result := Trunc(dwBytes / BytesPerSecond * 1000);
end;

// PʊZF~boCg
function TAuSong.MSecsToBytes(const dwMSecs: DWord): DWord;
begin
	Result := Ceil(dwMSecs / 1000 * BytesPerSecond);
end;

// ȏI[tO𗧂Ă
procedure TAuSong.SetEndFlag;
begin
	m_EndFlag := true;
	if Assigned(FOnSongEnd) then FOnSongEnd(Self);
end;

// ȏI[tO|
procedure TAuSong.ResetEndFlag;
begin
	m_EndFlag := false;
end;

function TAuSong.GetPluginFileName: TFileName;
begin
	Result := m_Plugin.FileName;
end;

// t@Cɂ̋Ȕԍݒ
// SeekTo() ƕpsB
procedure TAuSong.SetSongNum(const NewSongNum: DWord);
begin
	m_Plugin.SetPosition(m_hKMP, NewSongNum * 1000);
	ResetEndFlag;
	m_dwPosition := 0;
	m_dwSongNum := NewSongNum;
end;

function TAuSong.GetSoundInfo: PSOUNDINFO;
begin
	Result := @m_SoundInfo;
end;

end.
