unit kpiDriver;
{
 * NokomoDecoder
 * Copyright (c) 2005 Autch.net. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 *     Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 *     Neither the name of the Autch.net nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}

interface

uses Windows, SysUtils, Classes, Contnrs, KPI, KPIWrapper, KpiSong;

const
	PLUGIN_CACHE_VERSION	= 1;

type
	// false ԂƂ̃vOC[h͂邪 Enabled := false B
	TKpiLoadProgressEvent = function(Plugin: TAuKpi): Boolean of object;
	TKpiLoadFailedEvent = procedure(const FileName: TFileName) of object;

	// vOChCo
  //  TAuKpi ̔zAȃt@CK؂ TAuKpi I
  // TAuSong ɃbvĕԂB
	TAuKpiDriver = class(TObject)
		private
			procedure AddPlugin(const filename: TFileName);
			procedure SearchPlugins(const PluginPath: TFileName; const Depth: DWord);
			function GetPluginCount: integer;
			function CreateSongSub(const filename: TFileName; const pInfo: PSOUNDINFO): TAuSong;
			procedure SavePluginCache;
		protected
			m_Plugins: TObjectList;
			FOnPluginLoadProgress: TKpiLoadProgressEvent;
			FOnPluginLoadFailed: TKpiLoadFailedEvent;
			m_Extensions: TStringList;
		public
			constructor Create;
			procedure AddPlugins(const PluginPath: TFileName);
			function CreateSong(const filename: TFileName): TAuSong; overload;
			function CreateSong(const filename: TFileName; const pInfo: PSOUNDINFO): TAuSong; overload;
			destructor Destroy; override;

			property Count: integer	read GetPluginCount;
			property Plugins: TObjectList read m_Plugins;
			property OnPluginLoadProgress: TKpiLoadProgressEvent read FOnPluginLoadProgress write FOnPluginLoadProgress;
			property OnPluginLoadFailed: TKpiLoadFailedEvent read FOnPluginLoadFailed write FOnPluginLoadFailed;
			property Extensions: TStringList read m_Extensions;
	end;

implementation

uses Forms, StrUtils;

constructor TAuKpiDriver.Create;
begin
	inherited Create;

	m_Plugins := TObjectList.Create(true);		// Plugins  item L
	m_Plugins.Clear;
	m_Extensions := TStringList.Create;
	m_Extensions.CaseSensitive := false;
	m_Extensions.Sorted := true;
	m_Extensions.Duplicates := dupIgnore;
end;

// pluginpath ȉ̃fBNgXLČvOCo^
procedure TAuKpiDriver.AddPlugins(const PluginPath: TFileName);
begin
	SearchPlugins(PluginPath, $ff);
	m_Plugins.Sort(@CompareTAuKpi);
end;

destructor TAuKpiDriver.Destroy;
begin
	SavePluginCache;

	m_Plugins.Free;
	m_Extensions.Free;
	inherited;
end;

// KMPMODULE::Open() ̊e탉bp[
function TAuKpiDriver.CreateSong(const filename: TFileName): TAuSong;
begin
	Result := CreateSongSub(filename, nil);
end;

function TAuKpiDriver.CreateSong(const filename: TFileName; const pInfo: PSOUNDINFO): TAuSong;
begin
	Result := CreateSongSub(filename, pInfo);
end;

function TAuKpiDriver.CreateSongSub(const filename: TFileName; const pInfo: PSOUNDINFO): TAuSong;
var
	index: integer;
	item: TAuKpi;
	theExt: TFileName;
	foundIndex: integer;
	theSong: TAuSong;
begin
	theExt := ExtractFileExt(filename);
	theExt := MidStr(theExt, 2, StrLen(PChar(theExt)));
	for index := 0 to Plugins.Count - 1 do
	begin
    // evOC̑Ήgq𒲂ׂ
		item := (Plugins.items[index] as TAuKpi);
		if Assigned(item) = false then Continue; // nil ͓ĂȂ͂ǈꉞ
		if (item.Enabled = false) then Continue; // ݒɂ薳ɂĂ̂ŃX[
		foundIndex := 0;
		if (item.SupportedExts.Find(theExt, foundIndex) = true) then
		begin
			if Assigned(pInfo) then // SOUNDINFO w肳Ă΂ɏ]
				theSong := TAuSong.Create(filename, item, pInfo)
			else
				theSong := TAuSong.Create(filename, item);
			if Assigned(theSong) then
			begin
				Result := theSong;
				Exit;
			end;
      // vOC nil Ԃق̃vOC
		end;
	end;
  // gvOCȂ
	Result := nil;
end;

// P̃vOCJēo^
procedure TAuKpiDriver.AddPlugin(const filename: TFileName);
var
	item: TAuKpi;
begin
	try
		item := TAuKpi.Create(filename);
		if Assigned(FOnPluginLoadProgress) then
			item.Enabled := FOnPluginLoadProgress(item);
		Plugins.Add(item);

		m_Extensions.AddStrings(item.SupportedExts);

	except
		if Assigned(FOnPluginLoadFailed) then
			FOnPluginLoadFailed(filename);
	end;
end;

// pluginpath ȉ dir ɂvOCċAIɓo^܂Bdepth  dir ̊KwB
procedure TAuKpiDriver.SearchPlugins(const PluginPath: TFileName; const Depth: DWord);
var
	SearchRec: TSearchRec;
	Target: TFileName;
begin
	SetCurrentDir(pluginPath);
	// vOČ
	if (FindFirst(IncludeTrailingPathDelimiter(PluginPath) + '*.kpi', faReadOnly or faArchive, SearchRec) = 0) then
	begin
		repeat
			Target := IncludeTrailingPathDelimiter(PluginPath) + SearchRec.Name;
			if (SearchRec.Attr and faDirectory) = 0 then
				AddPlugin(target);
		until FindNext(SearchRec) <> 0;
		FindClose(SearchRec);
	end;

	// TufBNǧ
	if (Depth = 0) then Exit;
	if (FindFirst(IncludeTrailingPathDelimiter(PluginPath) + '*.*', faDirectory or faReadOnly or faArchive, SearchRec) = 0) then
	begin
		repeat
			Target := IncludeTrailingPathDelimiter(PluginPath) + SearchRec.Name;
			if ((SearchRec.Attr and faDirectory) <> 0) and	(SearchRec.Name <> '.') and (SearchRec.Name <> '..') then
				SearchPlugins(Target, Depth - 1);
		until FindNext(SearchRec) <> 0;
		FindClose(SearchRec);
	end;
end;

function TAuKpiDriver.GetPluginCount: integer;
begin
	Result := Plugins.Count;
end;

// vOCLbVۑ - ꉞǖgp
procedure TAuKpiDriver.SavePluginCache;
var
	CacheFile: TFileStream;
	i, t: DWord;
begin
	CacheFile := TFileStream.Create(ExtractFilePath(Application.ExeName) + 'plugin.cache', fmCreate);
	try
		with CacheFile do
		begin
			t := PLUGIN_CACHE_VERSION;
			Write(t, sizeof(DWord));
			Write(Plugins.Count, sizeof(DWord));

			for i := 0 to Plugins.Count - 1 do
				(Plugins.Items[i] as TAuKpi).WriteToStream(CacheFile);
		end;
	finally
		CacheFile.Free;
	end;
end;

end.
