unit kpiEffector;
{
 * NokomoDecoder
 * Copyright (c) 2005 Autch.net. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 *     Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 *     Neither the name of the Autch.net nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}

// GtFN^t[[NB
// fR[h PCM ɔCӂ̏{ƂłB
// ̂ƂĂ̂̓tF[_[B

interface

uses Windows, SysUtils, Classes, KPI, kpiWrapper, kpiSong, kpiDriver;

type
	TfnEffectorFunc = function(Song: TAuSong; const pBuffer: Pointer; const dwSize: DWord): DWord of object;

	// hăGtFN^
  // 2005.11.19:  interface Ŏׂ
	TAuKpiEffector = class(TObject)
		public
			function DoEffect(Song: TAuSong; const pBuffer: Pointer; const dwSize: DWord): DWord; virtual; abstract;
	end;

	TAuKpiFader = class(TAuKpiEffector)
		private
			// 1 sample iS`lΏہj ie. DoFade16() ̌Ői߂ׂ 2 * channels oCg
			procedure DoFade16(Song: TAuSong; const pBuffer: Pointer);
			procedure DoFade24(Song: TAuSong; const pBuffer: Pointer);
			procedure DoFade32(Song: TAuSong; const pBuffer: Pointer);
		protected
			m_dwFadeOutStart, m_dwFadeOutLength: DWord;
			m_dwVolumeRatio: DWord;
			m_dwDeltaVolRatio: DWord;		// VolRatio ̏l $ffffffff
		public
			constructor Create(dwFadeOutStart, dwFadeOutLength: DWord);
			function DoEffect(Song: TAuSong; const pBuffer: Pointer; const dwSize: DWord): DWord; override;
	end;

	// ȉ
	TAuKpiDigitalDistorter = class(TAuKpiEffector)
		private
			procedure DoFade16(Song: TAuSong; const pBuffer: Pointer);
			procedure DoFade24(Song: TAuSong; const pBuffer: Pointer);
			procedure DoFade32(Song: TAuSong; const pBuffer: Pointer);
		protected
			m_dwDisableBits: DWord;
		public
			function DoEffect(Song: TAuSong; const pBuffer: Pointer; const dwSize: DWord): DWord; override;
			property DisableBits: DWord read m_dwDisableBits write m_dwDisableBits;
	end;

	TAuKpiDigitalLoFi = class(TAuKpiEffector)
		private
			procedure DoEffect16(Song: TAuSong; const pBuffer: Pointer);
			procedure DoEffect24(Song: TAuSong; const pBuffer: Pointer);
			procedure DoEffect32(Song: TAuSong; const pBuffer: Pointer);
		protected
			m_dwDisableBits: DWord;
		public
			function DoEffect(Song: TAuSong; const pBuffer: Pointer; const dwSize: DWord): DWord; override;
			property DisableBits: DWord read m_dwDisableBits write m_dwDisableBits;
	end;

	TAuKpiDigitalEmphasis = class(TAuKpiEffector)
		private
			procedure DoEffect16(Song: TAuSong; const pBuffer: Pointer);
			procedure DoEffect24(Song: TAuSong; const pBuffer: Pointer);
			procedure DoEffect32(Song: TAuSong; const pBuffer: Pointer);
		protected
			m_dwDisableBits: DWord;
		public
			function DoEffect(Song: TAuSong; const pBuffer: Pointer; const dwSize: DWord): DWord; override;
			property DisableBits: DWord read m_dwDisableBits write m_dwDisableBits;
	end;
implementation

function TAuKpiFader.DoEffect(Song: TAuSong; const pBuffer: Pointer; const dwSize: DWord): DWord;
var
	P: PByte;
	dwBytesProcessed: DWord;
	fnFader: procedure(Song: TAuSong; const pBuffer: Pointer) of object;
begin
  // tF[hAEgԂɒBĂ邩H
	if Song.Position >= m_dwFadeOutStart then
		m_dwDeltaVolRatio := $ffffffff div Trunc(Song.SamplesPerSec * (m_dwFadeOutLength / 1000))
	else
	begin
    // ܂Ȃ牽ȂŋA
		Result := dwSize;
		Exit;
	end;

	case Song.BitsPerSample of
		16: fnFader := DoFade16;
		24: fnFader := DoFade24;
		32: fnFader := DoFade32;
	else
		raise Exception.Create('Unexpected sample bit-depth');
	end;

	P := PByte(pBuffer);
	dwBytesProcessed := 0;
	repeat
    // 1 sample * channels 
		fnFader(Song, P);

    // |C^i߂
		Inc(P, (Song.BitsPerSample div 8) * Song.Channels);
		Inc(dwBytesProcessed, (Song.BitsPerSample div 8) * Song.Channels);

    // tF[hAEg[g̒
		if m_dwDeltaVolRatio > 0 then
			if (m_dwVolumeRatio < m_dwDeltaVolRatio) then
				m_dwVolumeRatio := 0
			else
				Dec(m_dwVolumeRatio, m_dwDeltaVolRatio);
	until (m_dwVolumeRatio = 0) or (dwBytesProcessed = dwSize);

	Result := dwBytesProcessed;
end;

constructor TAuKpiFader.Create(dwFadeOutStart, dwFadeOutLength: DWord);
begin
	inherited Create;

	m_dwFadeOutStart := dwFadeOutStart;
	m_dwFadeOutLength := dwFadeOutLength;
	m_dwVolumeRatio := High(m_dwVolumeRatio);
end;

// 16bit * channels 
procedure TAuKpiFader.DoFade16(Song: TAuSong; const pBuffer: Pointer);
var
	P: PSmallInt;
	S: SmallInt;
	T: Int64;
	i: integer;
begin
	P := PSmallInt(pBuffer);
	for i := 1 to Song.Channels do
	begin
		S := P^;
		T := Int64(S) * Int64(m_dwVolumeRatio);
		S := SmallInt((T shr 32) and $ffff);
		//S := MulDiv(S, m_dwVolumeRatio, (1 shl 32));
		P^ := S;
		Inc(P);
	end;
end;

// 24bit * channels 
procedure TAuKpiFader.DoFade24(Song: TAuSong; const pBuffer: Pointer);
var
	P, P2: PByte;
	S: DWord;
	T: Int64;
	i: integer;
begin
	P := PByte(pBuffer);
	for i := 1 to Song.Channels do
	begin
		P2 := P;

		// ('A`)޸
    // 2005.11.19: 24bit ȂĎdl͑ȂȂĂ܂΂̂
		S := P2^;	Inc(P2); S := S or (P2^ shl 8); Inc(P2); S := S or (P2^ shl 16); Inc(P2);

		T := Int64(S) * Int64(m_dwVolumeRatio);
		S := (T shr 32) and $00ffffff;
		P^ := S and $ff;	Inc(P);	P^ := (S shr 8) and $ff;	Inc(P);	P^ := (S shr 16) and $ff;	Inc(P);

	end;
end;

// 32bit * channels 
procedure TAuKpiFader.DoFade32(Song: TAuSong; const pBuffer: Pointer);
var
	P: PLongInt;
	S: LongInt;
	T: Int64;
	i: integer;
begin
	P := PLongInt(pBuffer);
	for i := 1 to Song.Channels do
	begin
		S := P^;
		T := Int64(S) * Int64(m_dwVolumeRatio);
		S := LongInt((T shr 32) and $ffffffff);
		P^ := S;
		Inc(P);
	end;
end;

// ȉ TAuKpiFader Ɠl

function TAuKpiDigitalDistorter.DoEffect(Song: TAuSong; const pBuffer: Pointer; const dwSize: DWord): DWord;
var
	P: PByte;
	dwBytesProcessed: DWord;
	fnFader: procedure(Song: TAuSong; const pBuffer: Pointer) of object;
begin
	case Song.BitsPerSample of
		16: fnFader := DoFade16;
		24: fnFader := DoFade24;
		32: fnFader := DoFade32;
	else
		raise Exception.Create('Unexpected sample bit-depth');
	end;

	P := PByte(pBuffer);
	dwBytesProcessed := 0;
	while (dwBytesProcessed < dwSize) do
	begin
		fnFader(Song, P);

		Inc(P, (Song.BitsPerSample div 8) * Song.Channels);
		Inc(dwBytesProcessed, (Song.BitsPerSample div 8) * Song.Channels);
	end;

	Result := dwBytesProcessed;
end;

procedure TAuKpiDigitalDistorter.DoFade16(Song: TAuSong; const pBuffer: Pointer);
var
	P: PSmallInt;
	S: SmallInt;
	T: Int64;
	i: integer;
begin
	P := PSmallInt(pBuffer);
	for i := 1 to Song.Channels do
	begin
		S := P^;
		T := Int64(S) and not Int64(1 shl (Song.BitsPerSample - m_dwDisableBits));
		S := SmallInt(T and $ffff);
		P^ := S;
		Inc(P);
	end;
end;

procedure TAuKpiDigitalDistorter.DoFade24(Song: TAuSong; const pBuffer: Pointer);
var
	P, P2: PByte;
	S: DWord;
	T: Int64;
	i: integer;
begin
	P := PByte(pBuffer);
	for i := 1 to Song.Channels do
	begin
		P2 := P;

		// ('A`)޸
		S := P2^;	Inc(P2); S := S or (P2^ shl 8); Inc(P2); S := S or (P2^ shl 16); Inc(P2);

		T := Int64(S) and not Int64(1 shl (Song.BitsPerSample - m_dwDisableBits));
		S := T and $00ffffff;
		P^ := S and $ff;	Inc(P);	P^ := (S shr 8) and $ff;	Inc(P);	P^ := (S shr 16) and $ff;	Inc(P);
	end;
end;

procedure TAuKpiDigitalDistorter.DoFade32(Song: TAuSong; const pBuffer: Pointer);
var
	P: PLongInt;
	S: LongInt;
	T: Int64;
	i: integer;
begin
	P := PLongInt(pBuffer);
	for i := 1 to Song.Channels do
	begin
		S := P^;
		T := Int64(S) and not Int64(1 shl (Song.BitsPerSample - m_dwDisableBits));
		S := LongInt(T and $ffffffff);
		P^ := S;
		Inc(P);
	end;
end;

function TAuKpiDigitalLoFi.DoEffect(Song: TAuSong; const pBuffer: Pointer; const dwSize: DWord): DWord;
var
	P: PByte;
	dwBytesProcessed: DWord;
	fnFader: procedure(Song: TAuSong; const pBuffer: Pointer) of object;
begin
	case Song.BitsPerSample of
		16: fnFader := DoEffect16;
		24: fnFader := DoEffect24;
		32: fnFader := DoEffect32;
	else
		raise Exception.Create('Unexpected sample bit-depth');
	end;

	P := PByte(pBuffer);
	dwBytesProcessed := 0;
	while (dwBytesProcessed < dwSize) do
	begin
		fnFader(Song, P);

		Inc(P, (Song.BitsPerSample div 8) * Song.Channels);
		Inc(dwBytesProcessed, (Song.BitsPerSample div 8) * Song.Channels);
	end;

	Result := dwBytesProcessed;
end;

procedure TAuKpiDigitalLoFi.DoEffect16(Song: TAuSong; const pBuffer: Pointer);
var
	P: PSmallInt;
	S: SmallInt;
	T: Int64;
	i: integer;
begin
	P := PSmallInt(pBuffer);
	for i := 1 to Song.Channels do
	begin
		S := P^;
		T := DWord(S) and not ((1 shl m_dwDisableBits) - 1);
		S := SmallInt(T and $ffff);
		P^ := S;
		Inc(P);
	end;
end;

procedure TAuKpiDigitalLoFi.DoEffect24(Song: TAuSong; const pBuffer: Pointer);
var
	P, P2: PByte;
	S: DWord;
	T: Int64;
	i: integer;
begin
	P := PByte(pBuffer);
	for i := 1 to Song.Channels do
	begin
		P2 := P;

		// ('A`)޸
		S := P2^;	Inc(P2); S := S or (P2^ shl 8); Inc(P2); S := S or (P2^ shl 16); Inc(P2);

		T := DWord(S) and not ((1 shl m_dwDisableBits) - 1);
		S := T and $00ffffff;
		P^ := S and $ff;	Inc(P);	P^ := (S shr 8) and $ff;	Inc(P);	P^ := (S shr 16) and $ff; Inc(P);
	end;
end;

procedure TAuKpiDigitalLoFi.DoEffect32(Song: TAuSong; const pBuffer: Pointer);
var
	P: PLongInt;
	S: LongInt;
	T: Int64;
	i: integer;
begin
	P := PLongInt(pBuffer);
	for i := 1 to Song.Channels do
	begin
		S := P^;
		T := DWord(S) and not ((1 shl m_dwDisableBits) - 1);
		S := LongInt(T and $ffffffff);
		P^ := S;
		Inc(P);
	end;
end;

// 16bit ĂȂ

function TAuKpiDigitalEmphasis.DoEffect(Song: TAuSong; const pBuffer: Pointer; const dwSize: DWord): DWord;
var
	P: PByte;
	dwBytesProcessed: DWord;
	fnFader: procedure(Song: TAuSong; const pBuffer: Pointer) of object;
begin
	case Song.BitsPerSample of
		16: fnFader := DoEffect16;
		24: fnFader := DoEffect24;
		32: fnFader := DoEffect32;
	else
		raise Exception.Create('Unexpected sample bit-depth');
	end;

	P := PByte(pBuffer);
	dwBytesProcessed := 0;
	while (dwBytesProcessed < dwSize) do
	begin
		fnFader(Song, P);

		Inc(P, (Song.BitsPerSample div 8) * Song.Channels);
		Inc(dwBytesProcessed, (Song.BitsPerSample div 8) * Song.Channels);
	end;

	Result := dwBytesProcessed;
end;

procedure TAuKpiDigitalEmphasis.DoEffect16(Song: TAuSong; const pBuffer: Pointer);
var
	P: PSmallInt;
	S: SmallInt;
  ds: DWord;
	T: Int64;
	i: integer;
begin
	P := PSmallInt(pBuffer);
	for i := 1 to Song.Channels do
	begin
		S := P^;
    ds := S;
		T := -1 * ds * ds shr 15 + ds shl 1;  // t = -(ds^2 >> 15) + 2ds
		S := SmallInt(T and $ffff);
		P^ := S;
		Inc(P);
	end;
end;

procedure TAuKpiDigitalEmphasis.DoEffect24(Song: TAuSong; const pBuffer: Pointer);
var
	P, P2: PByte;
	S: DWord;
	T: Int64;
	i: integer;
begin
	P := PByte(pBuffer);
	for i := 1 to Song.Channels do
	begin
		P2 := P;

		// ('A`)޸
		S := P2^;	Inc(P2); S := S or (P2^ shl 8); Inc(P2); S := S or (P2^ shl 16); Inc(P2);

		T := DWord(S) and not ((1 shl m_dwDisableBits) - 1);
		S := T and $00ffffff;
		P^ := S and $ff;	Inc(P);	P^ := (S shr 8) and $ff;	Inc(P);	P^ := (S shr 16) and $ff; Inc(P);
	end;
end;

procedure TAuKpiDigitalEmphasis.DoEffect32(Song: TAuSong; const pBuffer: Pointer);
var
	P: PLongInt;
	S: LongInt;
	T: Int64;
	i: integer;
begin
	P := PLongInt(pBuffer);
	for i := 1 to Song.Channels do
	begin
		S := P^;
		T := DWord(S) and not ((1 shl m_dwDisableBits) - 1);
		S := LongInt(T and $ffffffff);
		P^ := S;
		Inc(P);
	end;
end;
end.
