unit  kpiPlayListReader;
{
 * NokomoDecoder
 * Copyright (c) 2005 Autch.net. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 *     Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 *     Neither the name of the Autch.net nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}

interface

uses SysUtils, Windows, Classes, StrUtils;

type
  TAuKbmPlayListTimeTagFormat = (ttfNone, ttfTxt, ttfLrc);
  // *.kbm `vCXg̋ȖڂЂƂ
  TAuKbmPlayListEntry = record
    sFileName: string;
    sNameInArchive: string;
    sTitle: string;
    sArtist:string;
    dwLength: DWord;
    sFileType: string;
    bHasMimpiWrd: boolean;
    bHasSherryWrd: boolean;
    eHasTimeTaggedLyrics: TAuKbmPlayListTimeTagFormat;
    bHasNeoWrd: boolean;
  end;


  // *.kbm `vCXgǂŋȖڂЂƂԂNX
  TAuKbmPlayListReader = class(TObject)
  protected
    m_slLines: TStringList;
    m_sFileName: string;
    m_Entry: TAuKbmPlayListEntry;
    m_dwIndex: DWord;
    m_bEndFlag: boolean;

    procedure FindNextEntry;
  public
    constructor Create; overload;
    constructor Create(const sKbmFile: string); overload;
    destructor Destroy; override;
    function LoadKbmFile(const sKbmFile: string): Boolean;
    function GetEntry: TAuKbmPlayListEntry;
    property IsEOF: boolean read m_bEndFlag;
  end;

implementation

constructor TAuKbmPlayListReader.Create;
begin
  m_slLines := nil;
  m_sFileName := '';
end;

constructor TAuKbmPlayListReader.Create(const sKbmFile: string);
begin
  if not LoadKbmFile(sKbmFile) then
    raise Exception.Create('G[FKBM vCXgJ܂B');
end;

destructor TAuKbmPlayListReader.Destroy;
begin
  if Assigned(m_slLines) then FreeAndNil(m_slLines);
end;

function TAuKbmPlayListReader.LoadKbmFile(const sKbmFile: string): Boolean;
begin
  Result := true;
  try
    m_slLines := TStringList.Create;
    m_slLines.LoadFromFile(sKbmFile);
    m_sFileName := sKbmFile;
    m_bEndFlag := false;
    m_dwIndex := 0;
    if (m_slLines[0] <> 'KbMIDI Player Play List') then
      Result := false;
  except
    Result := false;
  end;
  FindNextEntry;
end;

procedure TAuKbmPlayListReader.FindNextEntry;
var
  theLine: string;
  topChar: Char;
  function CutIfNotNull(sSource: string): string;
  begin
    if Length(sSource) > 2 then
      Result := Copy(sSource, 3, Length(sSource))
    else
      Result := '';
  end;
begin
  if m_dwIndex = m_slLines.Count then
  begin
    m_bEndFlag := true;
    Exit;
  end;

  with m_Entry do
  begin
    sFileName := '';
    sNameInArchive := '';
    sTitle := '';
    sArtist := '';
    dwLength := 0;
    sFileType := '';
    bHasMimpiWrd := false;
    bHasSherryWrd := false;
    eHasTimeTaggedLyrics := ttfNone;
    bHasNeoWrd := false;
  end;

  while m_dwIndex < m_slLines.Count do
  begin
    theLine := m_slLines[m_dwIndex];
    Inc(m_dwIndex);

    // sуRg̓X[
    if (theLine = '') or (theLine[1] = ';') then
      Continue;

    // ȕI[
    if Length(theLine) = 1 then Break;

    topChar := theLine[1];
    if topChar in ['0'..'9'] then
    begin
      case topChar of
        '0': m_Entry.sFileName := CutIfNotNull(theLine);
        '1': m_Entry.sNameInArchive := CutIfNotNull(theLine);
        '2': m_Entry.sTitle := CutIfNotNull(theLine);
        '3': m_Entry.sArtist := CutIfNotNull(theLine);
        // '4': m_Entry.dwLength := DWord(StrToInt(CutIfNotNull(theLine)));@// ǂfR[_ł͂Ȃ
        // '5': m_Entry.sFileType := CutIfNotNull(theLine);
        // 6/7/8/9 ('A`)޸
      end;
    end;
  end;
end;

function TAuKbmPlayListReader.GetEntry: TAuKbmPlayListEntry;
begin
  Result := m_Entry;
  if not m_bEndFlag then FindNextEntry;
end;

end.
