unit kpiWrapper;
{
 * NokomoDecoder
 * Copyright (c) 2005 Autch.net. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 *     Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 *     Neither the name of the Autch.net nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}

interface

uses Windows, SysUtils, Classes, KPI;

type
	// kpi W[ЂƂ̃bp[
	TAuKpi = class(TObject)
		//  kpi ̎dl𖞂ĂȂ́ihKMP  NULL nƗƂjłC
		// sł悤ɂĂ
		private
			function GetCopyright: string;
			function GetDescription: string;
			function GetReentrant: Boolean;
			procedure LoadPlugin;
		protected
			m_bCached: boolean;							// true: Open() Ƀ[hKvAfalse: sv
			m_hModule: HMODULE;
			m_szModuleFileName: TFileName;
			m_kmpModule: KMPMODULE;
			m_slSupportedExts: TStringList;
			m_bEnabled: boolean;
			m_dwFileAge: integer;
			m_dwOpenedCount: integer;				// Open()  increment, Close()  decrement
																			// ꂪ 0 ɂȂ܂ŁCCX^XĂ͂ȂB
		public
			constructor Create(const szFilename: TFileName); overload;
			constructor Create; overload;
			destructor Destroy; override;

			function Open(const filename: TFileName; const pInfo: PSOUNDINFO): HKMP;
			function OpenFromBuffer(const Buffer: PByte; const dwSize: DWord; const pInfo: PSOUNDINFO): HKMP;
			procedure Close(const hKMP: HKMP);
			function Render(const hKMP: HKMP; const Buffer: PByte; const dwSize: DWord): DWord;
			function SetPosition(const hKMP: HKMP; const dwPos: DWord): DWord;

			procedure WriteToStream(Stream: TStream);
			procedure ReadFromStream(Stream: TStream);

			property FileName: 			TFileName 	read m_szModuleFileName write m_szModuleFileName;
			property Module: 				KMPModule 	read m_kmpModule;
			property Copyright: 		string 			read GetCopyright;
			property Description: 	string 			read GetDescription;
			property SupportedExts: TStringList read m_slSupportedExts  write m_slSupportedExts default nil;
			property OpenedCount: 	integer 		read m_dwOpenedCount default 0;
			property Reentrant: 		Boolean 		read GetReentrant default false;
			property PluginVersion: DWord 			read m_kmpModule.dwPluginVersion write m_kmpModule.dwPluginVersion default 0;

			property Enabled:				Boolean			read m_bEnabled					write m_bEnabled default true;
			property FileAge:				integer			read m_dwFileAge				write m_dwFileAge default 0;
	end;

	TKpiLoadFailedReason = (frLoad, frGetAddress, frGetModule, frModuleVersion,
													frNoExtensions);
	EKpiLoadFailed = class(Exception)
		private
			m_Reason: TKpiLoadFailedReason;
			function GetReason: TKpiLoadFailedReason;
		public
			constructor Create(const Reason: TKpiLoadFailedReason);
			property Reason: TKpiLoadFailedReason read GetReason;
	end;

	// r֐
  function CompareTAuKpi(Item1, Item2: Pointer): integer;

implementation

constructor EKpiLoadFailed.Create(const Reason: TKpiLoadFailedReason);
begin
	m_Reason := Reason;
end;

function EKpiLoadFailed.GetReason: TKpiLoadFailedReason;
begin
	Result := m_Reason;
end;

constructor TAuKpi.Create(const szFilename: TFileName);
begin
	inherited Create;

	m_szModuleFileName := szFileName;
	m_bCached := true;
	LoadPlugin;
end;

constructor TAuKpi.Create;
begin
	inherited Create;

	m_bCached := true;
end;

procedure TAuKpi.LoadPlugin;
var
	getModule: fnGetKMPModule;
	pModule: PKMPModule;
	pSE: PPChar;
begin
	if not m_bCached then Exit;

	if m_szModuleFileName = '' then
		raise Exception.Create('szFileName == nil');

  // vOC[hB
  // m_hModule := LoadLibrary(PChar(m_szModuleFileName));
	m_hModule := LoadLibraryEx(PChar(m_szModuleFileName), 0, LOAD_WITH_ALTERED_SEARCH_PATH);
	if (m_hModule = 0) then
		raise EKpiLoadFailed.Create(frLoad);

	try
    // t@C̃^CX^vFTAuKpiDriver g
		m_dwFileAge := SysUtils.FileAge(m_szModuleFileName);

    // kpi Gg|Cg擾
		getModule := fnGetKMPModule(GetProcAddress(m_hModule, SZ_KMP_GETMODULE));
		if (@getModule = nil) then
			raise EKpiLoadFailed.Create(frGetAddress);

    // Gg|CgR[
		pModule := getModule();
		if Assigned(pModule) = false then
			raise EKpiLoadFailed.Create(frGetModule);

		m_kmpModule := pModule^;
		if (m_kmpModule.dwVersion <> KMPMODULE_VERSION) then
			raise EKpiLoadFailed.Create(frModuleVersion);

    // gq擾ăXg
		if (m_kmpModule.ppszSupportExts <> nil) then
		begin
			m_slSupportedExts := TStringList.Create;
			m_slSupportedExts.CaseSensitive := false;
			pSE := m_kmpModule.ppszSupportExts;
			while (pSE^ <> nil) do
			begin
				if (pSE^)^ = '.' then m_slSupportedExts.Add(pSE^+1);
				Inc(pSE);
			end;
			m_slSupportedExts.Sort;													// Sorted = true  Add() ኱
		end
		else
			raise EKpiLoadFailed.Create(frNoExtensions);
	except
		FreeLibrary(m_hModule);														// O𐶐Ίm FreeLibrary() 
																											// finally ɏĂ͂ȂB
		raise;
	end;
	InterlockedExchange(m_dwOpenedCount, 0); // QƃJEgFOpen() ꂽ

  m_bCached := false;

	if (@m_kmpModule.Init <> nil) then
		m_kmpModule.Init;
end;

destructor TAuKpi.Destroy;
begin
	if not m_bCached then
	begin
		if (@m_kmpModule.DeInit <> nil) then
			m_kmpModule.DeInit;
		if (m_hModule <> 0) then
			FreeLibrary(m_hModule);
	end;
	m_slSupportedExts.Free;

	inherited;
end;

function TAuKpi.GetCopyright: string;
begin
	Result := m_kmpModule.pszCopyright;
end;

function TAuKpi.GetDescription: string;
begin
	Result := m_kmpModule.pszDescription;
end;

// KMPMODULE::Open() ̃bp[
function TAuKpi.Open(const filename: TFileName; const pInfo: PSOUNDINFO): HKMP;
begin
	if m_bCached then	LoadPlugin;

	if ((m_dwOpenedCount > 0) and (m_kmpModule.dwReentrant = false)) then
	begin
		Result := nil; Exit;
	end;

	if (pInfo = nil) then begin Result := nil; Exit; end;
	Result := m_kmpModule.Open(PChar(Filename), pInfo);
	if (Result = nil) then begin Result := nil; Exit; end;
	if ((pInfo.dwSamplesPerSec or pInfo.dwChannels
		or pInfo.dwBitsPerSample) = 0) then
	begin
		Self.Close(Result);
		Result := nil;
		Exit;
	end;
  m_kmpModule.SetPosition(Result, 0);
	InterlockedIncrement(m_dwOpenedCount);
end;

procedure TAuKpi.Close(const hKMP: HKMP);
begin
	if (hKMP = nil) then Exit;
	m_kmpModule.Close(hKMP);
	InterlockedDecrement(m_dwOpenedCount);
end;

function TAuKpi.GetReentrant: Boolean;
begin
	Result := m_kmpModule.dwReentrant;
end;

function TAuKpi.OpenFromBuffer(const Buffer: PByte; const dwSize: DWord; const pInfo: PSOUNDINFO): HKMP;
begin
	if m_bCached then	LoadPlugin;

	if ((m_dwOpenedCount > 0) and (m_kmpModule.dwReentrant = false))
		or (@m_kmpModule.OpenFromBuffer = nil) then
	begin
		Result := nil;
		Exit;
	end;

	if (pInfo = nil) or (Buffer = nil) or (dwSize = 0) then
	begin
		Result := nil;
		Exit;
	end;
	Result := m_kmpModule.OpenFromBuffer(Buffer, dwSize, pInfo);
	if (Result = nil) then Exit;
	if ((pInfo.dwSamplesPerSec or pInfo.dwChannels
		or pInfo.dwBitsPerSample) = 0) then
	begin
		Self.Close(Result);
		Result := nil;
    Exit;
	end;
  InterlockedIncrement(m_dwOpenedCount);
end;

function TAuKpi.Render(const hKMP: HKMP; const Buffer: PByte; const dwSize: DWord): DWord;
begin
	if (hKMP = nil) then begin Result := 0; Exit; end;
	Result := m_kmpModule.Render(hKMP, Buffer, dwSize);
end;

function TAuKpi.SetPosition(const hKMP: HKMP; const dwPos: DWord): DWord;
begin
	if (hKMP = nil) then begin Result := 0; Exit; end;
	Result := m_kmpModule.SetPosition(hKMP, dwPos);
end;

function CompareTAuKpi(Item1, Item2: Pointer): integer;
var
	ver1, ver2: int64;
	age1, age2: int64;
begin
	ver1 := TAuKpi(Item1).PluginVersion;
	ver2 := TAuKpi(Item2).PluginVersion;
	Result := ver2 - ver1;
	if Result = 0 then
	begin
		age1 := TAuKpi(Item1).FileAge;
		age2 := TAuKpi(Item2).FileAge;
		Result := age2 - age1;
	end;
end;

// vOCLbV̈ꍀڂfoBgpB
procedure TAuKpi.WriteToStream(Stream: TStream);
var
	t: Byte;
	sBuffer: array[0..255] of Char;
begin
	with Stream do
	begin
		Write(m_dwFileAge, sizeof(m_dwFileAge));
		Write(m_kmpModule.dwPluginVersion, sizeof(m_kmpModule.dwPluginVersion));
		Write(ByteBool(m_bEnabled), sizeof(ByteBool));

		StrLCopy(@sBuffer, PChar(m_szModuleFileName), 255);
		t := StrLen(sBuffer);
		Write(t, sizeof(t));
		Write(sBuffer, StrLen(sBuffer));

		m_slSupportedExts.Delimiter := '/';
		StrLCopy(@sBuffer, PChar(m_slSupportedExts.DelimitedText), 255);
		t := StrLen(sBuffer);
		Write(t, sizeof(t));
		Write(sBuffer, StrLen(sBuffer));
	end;
end;

// vOCLbV玩g\zBgp
procedure TAuKpi.ReadFromStream(Stream: TStream);
var
	t: Byte;
	sBuffer: array[0..255] of Char;
begin
	With Stream do
	begin
		Read(m_dwFileAge, sizeof(m_dwFileAge));
		Read(m_kmpModule.dwPluginVersion, sizeof(m_kmpModule.dwPluginVersion));
		Read(ByteBool(m_bEnabled), sizeof(ByteBool));

		Read(t, SizeOf(t));
		Read(sBuffer, t);
		m_szModuleFileName := sBuffer;

		Read(t, SizeOf(t));
		Read(sBuffer, t);
		m_slSupportedExts.Delimiter := '/';
		m_slSupportedExts.DelimitedText := sBuffer;
	end;

	m_bCached := true;
end;


end.
